;; Problem 23
;; 02 August 2002
;;
;; A perfect number is a number for which the sum of its proper
;; divisors is exactly equal to the number. For example, the sum of
;; the proper divisors of 28 would be 1 + 2 + 4 + 7 + 14 = 28, which
;; means that 28 is a perfect number.
;;
;; A number whose proper divisors are less than the number is called
;; deficient and a number whose proper divisors exceed the number is
;; called abundant.

;; As 12 is the smallest abundant number, 1 + 2 + 3 + 4 + 6 = 16, the
;; smallest number that can be written as the sum of two abundant
;; numbers is 24. By mathematical analysis, it can be shown that all
;; integers greater than 28123 can be written as the sum of two
;; abundant numbers. However, this upper limit cannot be reduced any
;; further by analysis even though it is known that the greatest
;; number that cannot be expressed as the sum of two abundant numbers
;; is less than this limit.

;; Find the sum of all the positive integers which cannot be written
;; as the sum of two abundant numbers.

(defmacro acollect (&body body)
  (let ((resv (gensym)))
    `(let ((it (list))
	   (,resv nil))
      (setf ,resv (progn ,@body))
      (values it ,resv))))

(declaim (inline factor))

(defun factor (num)
  (declare (optimize speed)
	   (fixnum num))
  (acollect
    (push 1 it)
    (loop for ff from 2 to (+ 1 (ash num -1)) ;; optimization, close enough to (ceiling (/ num 2))
	  do
	  (when (and (= 0 (mod num ff))
		     (not (member ff it :test #'=)))
	    (push ff it)))))

;(factor 6)
;(factor 28122)
;(factor 28123)
;(factor 28124)

(declaim (inline sum-factors))
(defun sum-factors (num)
  (declare (optimize speed)
	   (fixnum num))
  (apply #'+ (factor num)))

(defun is-perfect (num)
  (declare (optimize speed)
	   (fixnum num))
  (= num (sum-factors num)))

(assert (is-perfect 6))
(assert (is-perfect 28))

(defun is-deficient (num)
  (declare (optimize speed)
	   (fixnum num))
  (< (sum-factors num) num))

(defun is-abundant1 (num)
  (> (sum-factors num) num))

(defun is-abundant (num)
  (declare (optimize speed)
	   (fixnum num))
  (let ((sum (sum-factors num)))
    (declare (fixnum sum))
    (> sum num)))

;(time (loop for ii from 1 to 20000 do (is-abundant1 ii)))
;(time (loop for ii from 1 to 20000 do (is-abundant ii)))

(defun find-abundants (num)
  (declare (optimize speed)
	   (fixnum num))
  (acollect
    (loop for ii from 3 to num
	  do
	  (when (is-abundant ii)
	    (push ii it)))))

; (find-abundants 1000)

(defvar *abundant-nums* (time (reverse (find-abundants 28123))))

; (sum-factors 12)
; (is-abundant 12)

; (subseq *abundant-nums* 0 10)

;(length *abundant-nums*)
; (princ *abundant-nums*)

(defvar *abundant-hash*
  (let ((hr (make-hash-table)))
    (loop for num in *abundant-nums*
	  do
	  (setf (gethash num hr) t))
    hr))

(defun find-nums-not-expressable-as-abundant-sums (max)
  (declare (optimize speed)
	   (fixnum max))
  (acollect
    (loop for ii from 1 to max
	  do
	  (when (= 0 (mod ii 100))
	    (format t "."))
	  (let ((can-be nil))
	      (loop for left in *abundant-nums*
		    for rem = (- ii left) then (- ii left)
		    do
		    (when (gethash rem *abundant-hash*)
		      (format t "found:~a=(+ ~a ~a)~&" ii left rem)
		      (setf can-be t)
		      (return t))
		    (when (< ii left)
		      ;;(format t "early exit at ii:~a vs ~a~&" ii left)
		      (return nil)))
	      (cond (can-be
		     nil)
		    (t
		     (push ii it)))))))

; (subseq *abundant-nums* 0 10)

; (time (find-nums-not-expressable-as-abundant-sums 1000))

;(apply #'+ (find-nums-not-expressable-as-abundant-sums 28123))

